#include  "pass_cource_table.h"
#include  "field_predicate.h"
#include  <cmath>
#include  <cfloat>
#include  <algorithm>

using namespace std;

#include  "debugstream.h"
using Debug_Stream::dbg;


class  Ball_Kickable_Predicate : public Field_Predicate
{
public:
	virtual	~Ball_Kickable_Predicate(){}

	virtual	bool	operator() ( const Field_Recog_Interface &  f ) const
	{
		return( (f.ball().coordinate() - f.self().coordinate()).r()
			< f.sserver_param().kickable_distance() );
	}
};


const	double	Pass_Cource_Table::PASS_EVAL_THRESHOLD = 15.0;

//
// Connection_Table
//
Pass_Cource_Table::Connection_Table::Connection_Table()
	: player( MAX_PLAYER + 1 , false ) , goal( false ) , push( false ) ,
	  free_radius( 0.0 ) ,
	  evaluation( 0.0 )
{
}


#include  "autogen_act_single_put_to.h"
#include  "autogen_act_hard_kick_to.h"
#include  "autogen_act_find_teammate.h"
#include  "autogen_act_keep_ball.h"
#include  "autogen_act_shoot.h"
#include  "autogen_act_face_to.h"
#include  "autogen_act_stop_ball.h"


//
// Pass_Cource_Table
//
Pass_Cource_Table::Pass_Cource_Table()
	: table( MAX_PLAYER + 1 ) ,
	  result() ,
	  first_best_player( -1 ) ,
	  table_prepared( false ) , search_finished( false )
{
}

Pass_Cource_Table::~Pass_Cource_Table()
{
}

void   Pass_Cource_Table::prepare_table( const Field_Recog_Interface &  f )
	const
{
	if ( table_prepared )
	{
		return;
	}

	for ( int  from = 1  ;  from <= MAX_PLAYER  ;  from ++ )
	{
		Player_Reference	fp = f.teammate(from);

		table[from].goal = this -> shoot_check( f , fp );
		dbg << "table[" << from << "].goal = "
		    << table[from].goal << endl;

		table[from].push = this -> push_check( f , fp );
		dbg << "table[" << from << "].push = "
		    << table[from].push << endl;

		for ( int  to = 1  ;  to <= MAX_PLAYER  ;  to ++ )
		{
			table[from].player[to]
			    = this -> pass_check( f , fp , f.teammate(to) );

			dbg << "table[" << from << "].player[" << to << "] = "
			    << table[from].player[to] << endl;
		}

		table[from].free_radius
			= this -> free_radius_check( f , fp );
	}

	for ( int  from = 1  ;  from <= MAX_PLAYER  ;  from ++ )
	{
		table[from].evaluation = this -> static_evaluation( f , from );

		dbg << "table[" << from << "].evaluation = "
		    << table[from].evaluation << endl;
	}

	table_prepared = true;
}


const std::vector< Pass_Cource_Table::Connection_Table > &
	Pass_Cource_Table::get_table( const Field_Recog_Interface &  f ) const
{
	this -> prepare_table( f );

	return( this -> table );
}

Pass_Cource_Table::Cource_Selection
  Pass_Cource_Table::get_first_selection( const Field_Recog_Interface &  f )
	const
{
	this -> prepare_table( f );

	Cource_Selection	c;
	c.player_number = f.self().player_number();

	if ( this -> table[f.self().player_number()].goal == true )
	{
		c.type = Cource_Selection::Shoot;

		return( c );
	}

	if ( this -> table[f.self().player_number()].push == true )
	{
		c.type = Cource_Selection::Push;

		return( c );
	}

	this -> prepare_result( f ,
				RECURSIVE_LEVEL ,
				f.self().player_number() );

	c.player_number = this -> first_best_player;

	if ( this -> first_best_player == f.self().player_number() )
	{
		c.type = Cource_Selection::Hold;
		return( c );
	}
	else
	{
		if ( f.teammate(this -> first_best_player)
		     .coordinate().accuracy_check( 0 ) )
		{
			c.type = Cource_Selection::Pass;
			return( c );
		}
		else
		{
			c.type = Cource_Selection::Find_Pass_Receiver;
			return( c );
		}
	}
}

Pass_Cource_Table::Cource_Selection  Pass_Cource_Table::get_result
					  ( const Field_Recog_Interface &  f )
	const
{
	this -> prepare_result( f ,
				RECURSIVE_LEVEL ,
				f.self().player_number() );

	return( this -> result );
}

ref_count_ptr<Soccer_Action>  Pass_Cource_Table::get_action
					( const Field_Recog_Interface &  f )
	const
{
	this -> prepare_table( f );

	if ( this -> table[f.self().player_number()].goal == true )
	{
		cerr << f.current_time().main_step() << ": Shoot" << endl;

		return( new Act_Shoot( f ) );
	}

	if ( this -> table[f.self().player_number()].push == true )
	{
		cerr << f.current_time().main_step() << ": Push" << endl;

		ref_count_ptr<Command_Queue>	ret_q = new Command_Queue;

#if 0
		if ( (f.ball().coordinate() - f.self().coordinate()).x()
		     < 0.3 )
#else
		if ( 0 )
#endif
		{
			// kick to corner
			double	to_x
				= f.sserver_param().field_half_length() - 3.0;

			double	to_y_abs
				= f.sserver_param().field_half_width() - 10.0;

			D2_Vector	to;

			if ( f.self().coordinate().y() >= 0 )
			{
				to.set( to_x , + to_y_abs );
			}
			else
			{
				to.set( to_x , - to_y_abs );
			}

			ret_q -> push( new Act_Single_Put_To( f , to ) );
		}
		else
		{
			// drrible
			ret_q -> push( new Time_Limit_Queue
				       ( f , 1 , new Act_Keep_Ball( f ) ) );

#if 1
			ret_q -> push( new Act_Single_Put_To
				       ( f ,
					 f.self().coordinate()
					   + D2_Vector( 15.0 , 0.0 ) ) );
#else
			ret_q -> push( new Act_Single_Put_To
				       ( f ,
					 f.self().coordinate()
					   + D2_Vector( 35.0 , 0.0 ) ) );
#endif
		}

		return( ret_q );
	}


	this -> prepare_result( f ,
				RECURSIVE_LEVEL ,
				f.self().player_number() );

	if ( this -> first_best_player == f.self().player_number() )
	{
		cerr << f.current_time().main_step() << ": Keep_Ball" << endl;

		return( new Time_Limit_Queue
			( f , 2 , new Act_Keep_Ball( f ) ) );
	}
	else
	{
		if ( f.teammate( this -> first_best_player )
		     .coordinate().accuracy_check( 0 ) )
		{
			cerr << f.current_time().main_step()
			     << ": Pass(" << this -> first_best_player << ")"
			     << endl;

			// XXX
			f.set_target_teammate( this -> first_best_player );

			return( new Act_Hard_Kick_To
				( f ,
				  f.teammate( this -> first_best_player )
							  .coordinate()
				  + f.teammate( this -> first_best_player )
							  .velocity() ) );
		}
		else
		{
			cerr << f.current_time().main_step()
			     << ": Find_Teammate("
			     << this -> first_best_player << ")"
			     << endl;

			ref_count_ptr<Condition_Queue>
				ret_q = new Condition_Queue
					    ( f ,
					      new Ball_Kickable_Predicate );

			ret_q -> push( new Act_Stop_Ball( f ) );

			ret_q -> push( new Time_Limit_Queue
				       ( f , 3 ,
					 new Act_Find_Teammate
					 ( f , first_best_player ) ) );

#if 0
			// XXX
			f.set_pass_receiver( first_best_player );
#endif

			return( ret_q );
		}
	}
}


double  Pass_Cource_Table::pass_check( const Field_Recog_Interface &  f ,
				       const Player_Reference &  from ,
				       const Player_Reference &  to ) const
{
	const	long	VALID_TEAMMATE_ACCURACY = - 40;
	const	long	VALID_OPPONENT_ACCURACY = - 100;

	if ( ! from.coordinate().accuracy_check( VALID_TEAMMATE_ACCURACY )
	  || ! to  .coordinate().accuracy_check( VALID_TEAMMATE_ACCURACY ) )
	{
		return( 0.0 );
	}

	if ( (to.coordinate() - from.coordinate()).r() >= 45.0 )
	{
		dbg << "too far" << endl;
		return( 0.0 );
	}

	if ( to.coordinate().x() >= from.team().offence_line() + 3.0 )
	{
		dbg << "offside" << endl;
		return( 0.0 );
	}

	if ( to.coordinate().x()
	     <= f.sserver_param().field_half_length()
		- f.sserver_param().penalty_area_length()
		+ 10.0
	  && std::fabs( to.coordinate().y() )
	     <= f.sserver_param().penalty_area_half_width() + 5.0 )
	{
		dbg << "in penalty area" << endl;
		return( 0.0 );
	}

	if ( fabs( to.coordinate().x() )
	     >= f.sserver_param().field_half_length()
	  || fabs( to.coordinate().y() )
	     >= f.sserver_param().field_half_width() )
	{
		dbg << "out of field" << endl;
		return( 0.0 );
	}

	if ( to.goalie()
	  && to.coordinate().x() < - 36.0
	  && fabs( to.coordinate().y() ) < 20.16 )
	{
		dbg << "goalie can not catch" << endl;
		return( 0.0 );
	}


	D2_Vector	from_coordinate = from.coordinate();

	if ( from.self_player() )
	{
		from_coordinate = f.ball().coordinate();
	}

	Angle	test_dir = (to.coordinate() - from_coordinate)
			   .theta().normalize();

	Angle	pass_cource_cone = Degree( + 360.0 );

	for ( int  opp = 1  ;  opp <= MAX_PLAYER  ;  opp ++ )
	{
		Player_Reference	op = f.opponent(opp);

		if ( ! op.coordinate()
			      .accuracy_check( VALID_OPPONENT_ACCURACY ) )
		{
			continue;
		}

		if ( (op.coordinate() - from_coordinate).r() - 5.0
		   > (to.coordinate() - from_coordinate).r() )
		{
			continue;
		}

		Angle	angle_diff
			= ((op.coordinate() - from_coordinate).theta()
			   - test_dir).normalize().abs();

		if ( from.self_player() )
		{
			Radian	hide_angle
				( std::asin
				  ( std::min( f.sserver_param()
						      .kickable_distance()
						/ (op.coordinate()
						   - from_coordinate).r() ,
					      1.0 ) ) );
			angle_diff = std::max( angle_diff - hide_angle ,
					       Angle::origin() );
		}

		if ( pass_cource_cone > angle_diff )
		{
			pass_cource_cone = angle_diff;
		}
	}

	for ( size_t  i = 0  ;  i < f.n_unknown_opponent()  ;  i ++ )
	{
		Player_Reference	op = f.unknown_opponent(i);

		if ( ! op.coordinate()
			      .accuracy_check( VALID_OPPONENT_ACCURACY ) )
		{
			continue;
		}

		if ( (op.coordinate() - from_coordinate).r() - 5.0
		   > (to.coordinate() - from_coordinate).r() )
		{
			continue;
		}

		Angle	angle_diff
			= ((op.coordinate() - from_coordinate).theta()
			   - test_dir).normalize().abs();

		if ( from.self_player() )
		{
			Radian	hide_angle
				( std::asin
				  ( std::min( f.sserver_param()
						      .kickable_distance()
						/ (op.coordinate()
						   - from_coordinate).r() ,
					      1.0 ) ) );
			angle_diff = std::max( angle_diff - hide_angle ,
					       Angle::origin() );
		}

		if ( pass_cource_cone > angle_diff )
		{
			pass_cource_cone = angle_diff;
		}
	}

	for ( size_t  i = 0  ;  i < f.n_unknown_player()  ;  i ++ )
	{
		Player_Reference	op = f.unknown_player(i);

		if ( ! op.coordinate()
			      .accuracy_check( VALID_OPPONENT_ACCURACY ) )
		{
			continue;
		}

		if ( (op.coordinate() - from_coordinate).r() - 5.0
		   > (to.coordinate() - from_coordinate).r() )
		{
			continue;
		}

		Angle	angle_diff
			= ((op.coordinate() - from_coordinate).theta()
			   - test_dir).normalize().abs();

		if ( from.self_player() )
		{
			Radian	hide_angle
				( std::asin
				  ( std::min( f.sserver_param()
						      .kickable_distance()
						/ (op.coordinate()
						   - from_coordinate).r() ,
					      1.0 ) ) );
			angle_diff = std::max( angle_diff - hide_angle ,
					       Angle::origin() );
		}

		if ( pass_cource_cone > angle_diff )
		{
			pass_cource_cone = angle_diff;
		}
	}

	if ( pass_cource_cone > Degree( PASS_EVAL_THRESHOLD ) )
	{
		return( pass_cource_cone.degree() );
	}
	else
	{
		dbg << "pass cource too narrow: "
		    << pass_cource_cone.degree() << endl;

		return( 0.0 );
	}
}


double  Pass_Cource_Table::free_radius_check
				( const Field_Recog_Interface &  f ,
				  const Player_Reference &  pl ) const
{
	const	long	VALID_TEAMMATE_ACCURACY = - 40;
	const	long	VALID_OPPONENT_ACCURACY = - 100;

	if ( ! pl.coordinate().accuracy_check( VALID_TEAMMATE_ACCURACY ) )
	{
		return( 0.0 );
	}

	double	r_min = f.sserver_param().field_diagonal_length();

	for ( int  opp = 1  ;  opp <= MAX_PLAYER  ;  opp ++ )
	{
		Player_Reference	op = f.opponent(opp);

		if ( ! op.coordinate()
			      .accuracy_check( VALID_OPPONENT_ACCURACY ) )
		{
			continue;
		}

		double	r = (op.coordinate() - pl.coordinate()).r();

		if ( r < r_min )
		{
			r_min = r;
		}
	}

	return( r_min );
}


bool   Pass_Cource_Table::shoot_check( const Field_Recog_Interface &  f ,
				       const Player_Reference &  from ) const
{
	const	long	VALID_TEAMMATE_ACCURACY = - 40;
	const	long	VALID_OPPONENT_ACCURACY = - 40;
	const	long	VALID_FAR_OPPONENT_ACCURACY = - 200;

	if ( ! from.coordinate().accuracy_check( VALID_TEAMMATE_ACCURACY ) )
	{
		return( false );
	}

	if ( (from.coordinate()
	      - D2_Vector( f.sserver_param().field_half_length() , 0.0 )).r()
	     > 34.0 )
	{
		return( false );
	}


	long	acc = VALID_OPPONENT_ACCURACY;

	if ( from.coordinate().x() <= 30.0 )
	{
		acc = VALID_FAR_OPPONENT_ACCURACY;
	}

	return( Act_Shoot::free_angle( f , acc ) >= Degree( 15.0 ) );
}


bool   Pass_Cource_Table::push_check( const Field_Recog_Interface &  f ,
				      const Player_Reference &  from ) const
{
	const	long	VALID_PLAYER_ACCURACY = - 50;

	if ( ! from.coordinate().accuracy_check( VALID_PLAYER_ACCURACY )
	  || from.coordinate().x() < from.team().offence_line() - 10.0
	  || from.coordinate().x()
	     > f.sserver_param().field_half_length() - 10.0
	  || (from.coordinate().x()
	      > (f.sserver_param().field_half_length()
		 - f.sserver_param().penalty_area_length())
	     && std::fabs( from.coordinate().y() )
	      < f.sserver_param().field_half_width() - 5.0 ) )
	{
		return( false );
	}

	for ( int  opp = 1  ;  opp < MAX_PLAYER  ;  opp ++ )
	{
		if ( ! f.opponent(opp).coordinate()
		         .accuracy_check( VALID_PLAYER_ACCURACY ) )
		{
			continue;
		}

		if ( f.opponent(opp).coordinate().x() > from.coordinate().x()
		  && fabs(f.opponent(opp).coordinate().y()
			  - from.coordinate().y()) < 4.0 )
		{
			return( false );
		}
	}

	for ( size_t  i = 0  ;  i < f.n_unknown_opponent()  ;  i ++ )
	{
		if ( ! f.unknown_opponent(i).coordinate()
		         .accuracy_check( VALID_PLAYER_ACCURACY ) )
		{
			continue;
		}

		if ( f.unknown_opponent(i).coordinate().x()
		     > from.coordinate().x()
		 && fabs(f.unknown_opponent(i).coordinate().y()
			 - from.coordinate().y()) < 4.0 )
		{
			return( false );
		}
	}

	for ( size_t  i = 0  ;  i < f.n_unknown_player()  ;  i ++ )
	{
		if ( ! f.unknown_player(i).coordinate()
		         .accuracy_check( VALID_PLAYER_ACCURACY ) )
		{
			continue;
		}

		if ( f.unknown_player(i).coordinate().x()
		     > from.coordinate().x()
		  && fabs(f.unknown_player(i).coordinate().y()
			  - from.coordinate().y()) < 4.0 )
		{
			return( false );
		}
	}

	return( true );
}


void   Pass_Cource_Table::prepare_result( const Field_Recog_Interface &  f ,
					  int  recursive_level ,
					  int  from ) const
{
	if ( this -> search_finished )
	{
		return;
	}

	if ( table[from].goal )
	{
		this -> first_best_player = from;

		this -> result.type = Cource_Selection::Shoot;
		this -> result.player_number = from;
	}
	else if ( table[from].push )
	{
		this -> first_best_player = from;

		this -> result.type = Cource_Selection::Push;
		this -> result.player_number = from;
	}
	else if ( recursive_level == 0 )
	{
		this -> first_best_player = from;

		this -> result.type = Cource_Selection::Hold;
		this -> result.player_number = from;
	}


	double	max_eval = table[from].evaluation;
	this -> first_best_player = from;

	for ( int  i = 1  ;  i <= MAX_PLAYER  ;  i ++ )
	{
		if ( i == from
		  || this -> table[from].player[i] < PASS_EVAL_THRESHOLD )
		{
			continue;
		}

		Cource_Selection	r;
		double	ev = recursive_search( f , recursive_level - 1 , i ,
					       &r );

		if ( ev > max_eval
		  || ev == max_eval
		     && ((table[from].player[i] > PASS_EVAL_THRESHOLD
			  && table[from].player[this -> first_best_player]
			     > PASS_EVAL_THRESHOLD
			  && table[i].evaluation
			     > table[this -> first_best_player].evaluation) ) )
		{
			max_eval = ev;
			this -> first_best_player = i;
			this -> result = r;
		}
	}

	this -> search_finished = true;
}


double  Pass_Cource_Table::recursive_search( const Field_Recog_Interface &  f ,
					     int  recursive_level ,
					     int  from ,
					     Cource_Selection *  r )
	const
{
	if ( table[from].goal )
	{
		r -> type = Cource_Selection::Shoot;
		r -> player_number = from;

		return( table[from].evaluation );
	}
	else if ( table[from].push )
	{
		r -> type = Cource_Selection::Push;
		r -> player_number = from;

		return( table[from].evaluation );
	}
	else if ( recursive_level == 0 )
	{
		r -> type = Cource_Selection::Hold;
		r -> player_number = from;

		return( table[from].evaluation );
	}


	double	max_eval = table[from].evaluation;

	for ( int  i = 1  ;  i <= MAX_PLAYER  ;  i ++ )
	{
		if ( i == from
		  || this -> table[from].player[i] < PASS_EVAL_THRESHOLD )
		{
			continue;
		}

		Cource_Selection	serarch_result;

		double	ev = recursive_search( f , recursive_level - 1 , i ,
					       &serarch_result );

		if ( ev > max_eval )
		{
			max_eval = ev;
			*r = serarch_result;
		}
	}

	return( max_eval );
}


double  Pass_Cource_Table::static_evaluation
	  ( const Field_Recog_Interface &  f ,  int  player_number ) const
{
	double	point = f.teammate(player_number).coordinate().x();

	if ( table[player_number].goal == true )
	{
		point += 1000.0;
	}
	else if ( table[player_number].push == true )
	{
		point += 500.0;
	}

#if 0
	// XXX: cross ball
	if ( f.self().coordinate().x()
	     > - (f.sserver_param().field_half_length()
		  - f.sserver_param().penalty_area_length() ) + 10.0 )
	{
		point += fabs( f.teammate(player_number).coordinate().y()
			       -  f.self().coordinate().y() );
	}
#endif

	if ( f.teammate(player_number).coordinate().x()
	     > f.self().coordinate().x() )
	{
		point += table[player_number].free_radius;
	}

	return( point );
}
